#include-once

#cs ================================================================================================
2017-08-17 - Added Global Notify Handler 
#ce ================================================================================================

Global Enum $NH_MSG, $NH_FUNC, $NH_DATA, $NH_HWND, $NH_STATE, $NH_MAX
Global Enum $NH_STATE_FREE, $NH_STATE_NEW, $NH_STATE_ACTIVE
Global Enum $NH_DBKCLICK_CTRL, $NH_DBKCLICK_INVOKE
Global $aGlobalNotifyHandlerList[0][$NH_MAX] ; 

; sets default icon that designated to file
Func _TCH_GUI_SetDefaultIcon()
    If @Compiled Then Return
    Local $sIcon = IniRead(@ScriptFullPath, "Info", "Icon", "")
    Local $aIcon = StringRegExp($sIcon, "(.+?),(\d+)$", 1)
    If @error Then
        GUISetIcon($sIcon, 0)
    Else
        GUISetIcon(StringReplace($aIcon[0], "%COMMANDER_PATH%", EnvGet("COMMANDER_PATH")), -$aIcon[1]-1)
    EndIf
EndFunc

; relatively rough thing to correct gui size on the 120 dpi screens
; need Opt("GUIResizeMode", $GUI_DOCKAUTO) to be set before gui create
; also this functon should be placed AFTER GUISetState(@SW_SHOW)
Func _TCH_GUI_AdjustSize($hWnd)
    Local $aDPI = __GetDPI()
    If $aDPI[0] = 96 Then Return
    Local $iDPIRat = $aDPI[1]
    Local $pos = WinGetPos($hWnd)
    Local $relX =  (($pos[2] * $iDPIRat) - $pos[2]) / 2
    Local $relY =  (($pos[3] * $iDPIRat) - $pos[3]) / 2
    WinMove($hWnd, "", $pos[0] - $relX, $pos[1] - $relY, $pos[2] * $iDPIRat, $pos[3] * $iDPIRat)
EndFunc

; Show Popup menu at given coordinate
Func _TrackPopupMenu($hWnd, $hMenu, $x, $y)
    DllCall("user32.dll", "int", "TrackPopupMenuEx", "hwnd", $hMenu, "int", 0, "int", $x, "int", $y, "hwnd", $hWnd, "ptr", 0)
EndFunc

; Convert the client (GUI) coordinates to screen (desktop) coordinates
Func _ClientToScreen($hWnd, ByRef $x, ByRef $y)
    Local $stPoint = DllStructCreate("int;int")
    
    DllStructSetData($stPoint, 1, $x)
    DllStructSetData($stPoint, 2, $y)

    DllCall("user32.dll", "int", "ClientToScreen", "hwnd", $hWnd, "ptr", DllStructGetPtr($stPoint))
    
    $x = DllStructGetData($stPoint, 1)
    $y = DllStructGetData($stPoint, 2)
EndFunc

Func __GetDPI()
   ;; Author: Phillip123Adams
   ;; Modified by YDY
   Local $a1[2]
   Local $iDPI, $iDPIRat, $Logpixelsy = 90, $hWnd = 0
   Local $hDC = DllCall("user32.dll", "long", "GetDC", "long", $hWnd)
   Local $aRet = DllCall("gdi32.dll", "long", "GetDeviceCaps", "long", $hDC[0], "long", $Logpixelsy)
   Local $hNewDC = DllCall("user32.dll", "long", "ReleaseDC", "long", $hWnd, "long", $hDC[0])
   $iDPI = $aRet[0]
   ;; Set a ratio for the GUI dimensions based upon the current DPI value.
   Select
      Case $iDPI = 0
         $iDPI = 96
         $iDPIRat =  94
      Case $iDPI < 84
         $iDPIRat =  $iDPI / 105
      Case $iDPI < 121
         $iDPIRat =  $iDPI / 96
      Case $iDPI < 145
         $iDPIRat =  $iDPI / 95
      Case Else
         $iDPIRat =  $iDPI / 94
   EndSelect
   $a1[0] = $iDPI
   $a1[1] = $iDPIRat
   Return $a1
EndFunc

Func _TCH_GUI_MenuRightJustify($hGUI, $nItem, $isIndex = false)
    Local $MFT_RIGHTJUSTIFY = 0x4000
    Local $MIIM_TYPE = 0x10

    Local $tagMENUITEMINFO = "uint;uint;uint;uint;uint;uint;uint;uint;dword;ptr;uint"
    Local $menu = DllCall("user32.dll", "hwnd", "GetMenu", "hwnd", $hGUI)
    $menu = $menu[0]

    Local $menuinfo = DllStructCreate($tagMENUITEMINFO)
    Local $buffer   = DllStructCreate("byte[80]")

    DllStructSetData($menuinfo, 1, DllStructGetSize($menuinfo))
    DllStructSetData($menuinfo, 2, $MIIM_TYPE)
    DllStructSetData($menuinfo, 10, DllStructGetPtr($buffer))
    DllStructSetData($menuinfo, 11, 80)

    DllCall("user32.dll", "int", "GetMenuItemInfo", "int", $menu, _
                                                    "int", $nItem, _
                                                    "int", $isIndex, _
                                                    "ptr", DllStructGetPtr($menuinfo))

    DllStructSetData($menuinfo, 3, BitOR(DllStructGetData($menuinfo, 3), $MFT_RIGHTJUSTIFY))

    DllCall("user32.dll", "int", "SetMenuItemInfo", "int", $menu, _
                                                    "int", $nItem, _
                                                    "int", $isIndex, _
                                                    "ptr", DllStructGetPtr($menuinfo))
EndFunc


#cs
This set of functions is intended to handle GUI messages flow with easily variables 
exchange between caller and callee (message handler) functions. 

; adding message to handler
_TCHG_MessageHandlerAdd($nMessage, $fFunc, $vData = Null)

; shortcut for doubleclick as one of most used
; $hControl - control for which doubleclick will be registered
; $hInvoke - message will be send to this control in case of doubleclick
_TCHG_DoubleClickHandlerAdd($hControl, $hInvoke)

; register message for current GUI
_TCHG_MessageHandlerRegister($hWnd)

; unregister messsage for current GUI befor this GUI being deleted
_TCHG_MessageHandlerUnregister($vData)

; function that registered for message is almost the same as standard windows
; message, but adding one integer parameter before all other params, this is array index
; where message registered, it used for address of correct data
_TCHG_DoubleClickFunc($vKey, $hWnd, $MsgID, $wParam, $lParam)
#ce

Func _TCHG_MessageHandlerAdd($nMessage, $fFunc, $vData = Null)
    ; find element with "free" state
    Local $nIndex = -1
    For $i = 0 To UBound($aGlobalNotifyHandlerList) - 1
        If $aGlobalNotifyHandlerList[$i][$NH_STATE] = $NH_STATE_FREE Then
            $nIndex = $i
            ExitLoop
        EndIf
    Next
    ; if not found, add array element
    If $nIndex = -1 Then
        $nIndex = UBound($aGlobalNotifyHandlerList)
        ReDim $aGlobalNotifyHandlerList[$nIndex + 1][$NH_MAX]
    EndIf

    $aGlobalNotifyHandlerList[$nIndex][$NH_HWND]   = 0
    $aGlobalNotifyHandlerList[$nIndex][$NH_MSG]    = $nMessage
    $aGlobalNotifyHandlerList[$nIndex][$NH_FUNC]   = $fFunc
    $aGlobalNotifyHandlerList[$nIndex][$NH_STATE]  = $NH_STATE_NEW

    If @NumParams = 2 Then
      Local $mTemp[]
      $aGlobalNotifyHandlerList[$nIndex][$NH_DATA] = $mTemp
    Else
      $aGlobalNotifyHandlerList[$nIndex][$NH_DATA] = $vData
    EndIf 
    
    Return $nIndex
EndFunc

Func _TCHG_MessageHandlerRegister($hWnd)
    ; register lately added messages for current GUI
    For $i = 0 To UBound($aGlobalNotifyHandlerList) - 1
        If $aGlobalNotifyHandlerList[$i][$NH_STATE] = $NH_STATE_NEW Then 
            $aGlobalNotifyHandlerList[$i][$NH_HWND] = $hWnd
            GUIRegisterMsg($aGlobalNotifyHandlerList[$i][$NH_MSG], "_TCHG_NotifyHandlerFunc")
            $aGlobalNotifyHandlerList[$i][$NH_STATE] = $NH_STATE_ACTIVE
        EndIf 
    Next
EndFunc

; this function handle notifications of all types
Func _TCHG_NotifyHandlerFunc($hWnd, $MsgID, $wParam, $lParam)
    Local $nResult
    For $vKey = 0 To UBound($aGlobalNotifyHandlerList)-1
        ; don't not check for apriori incorrect windows
        If $aGlobalNotifyHandlerList[$vKey][$NH_STATE] <> $NH_STATE_ACTIVE  Then ContinueLoop
        If $aGlobalNotifyHandlerList[$vKey][$NH_MSG]   <> $MsgID Then ContinueLoop
        If $aGlobalNotifyHandlerList[$vKey][$NH_HWND]  <> $hWnd Then ContinueLoop
        ; execute assigned function
        $nResult = ($aGlobalNotifyHandlerList[$vKey][$NH_FUNC])($vKey, $hWnd, $MsgID, $wParam, $lParam)
        ; if message not handled, continue loop, else return result
        if $nResult = $GUI_RUNDEFMSG Then
            ContinueLoop
        Else
            Return $nResult
        EndIf
    Next
    Return $GUI_RUNDEFMSG
EndFunc

Func _TCHG_MessageHandlerUnregister($vData)
    If IsHWnd($vData) Then
        For $i = 0 To UBound($aGlobalNotifyHandlerList)-1
            If $aGlobalNotifyHandlerList[$i][$NH_HWND] = $vData Then
                $aGlobalNotifyHandlerList[$i][$NH_HWND] = 0
                $aGlobalNotifyHandlerList[$i][$NH_STATE] = $NH_STATE_FREE
            EndIf 
        Next
    Else
        $aGlobalNotifyHandlerList[$vData][$NH_STATE] = $NH_STATE_FREE
    EndIf 
EndFunc

; One of most useful handlers shortcut
; doubleclick handler
Func _TCHG_DoubleClickHandlerAdd($hControl, $hInvoke)
;    Local $mData[], 
    Local $aData[] = [$hControl, $hInvoke]
;    $mData.control = $hControl
;    $mData.invoke  = $hInvoke
    _TCHG_MessageHandlerAdd($WM_NOTIFY, _TCHG_DoubleClickFunc, $aData)
EndFunc

Func _TCHG_DoubleClickFunc($vKey, $hWnd, $MsgID, $wParam, $lParam)
    Local $tagNMHDR, $code
    $tagNMHDR = DllStructCreate("int;int;int", $lParam)
    If @error Then Return $GUI_RUNDEFMSG
    $code = DllStructGetData($tagNMHDR, 3)
    Local $aData = $aGlobalNotifyHandlerList[$vKey][$NH_DATA]
    If ($wParam = $aData[$NH_DBKCLICK_CTRL]) And ($code = $NM_DBLCLK) Then
        DllCall("user32.dll", "int", "PostMessage", "hwnd", $hWnd, "int", $WM_COMMAND, "wparam", $aData[$NH_DBKCLICK_INVOKE], "lparam", 0)
        Return ; handled
    EndIf 
    Return $GUI_RUNDEFMSG
EndFunc
