#include-once
;#include <Array.au3>
;#include <Common\Debug.au3>

#cs ================================================================================================
This file contain different Total Commander related variables and code
#ce ================================================================================================

#cs ================================================================================================
2020-06-09 - some cleanup
2020-05-01 - removed deprecated _TCHLoadFileList
2020-03-12 - moved Map-dependent code to TCHelpersMap.au3
2020-03-12 - redone translation code without maps
2020-03-11 - improved temporary files/dirs handling, added autodelete on exit
2018-04-25 - Added map helpers
2018-03-05 - _TCH_GetScriptInfo support for compiled scripts
2018-02-00 - Updated to latest changes in structure
2017-12-17 - Updated to latest changes in pack structure, fixed continious lists redirection awareness
2017-10-11 - Added temporary files handling functions
2017-03-21 - Fixed (again) _TCExtractFileName
2017-03-21 - Fixed wrong icon index in _TCHCreateButton function
2017-02-27 - Added history block, added _TCHSendMsg function, added _TCHLockWindowUpdate function
#ce ================================================================================================

; #CONSTANTS# ======================================================================================
Global Const $TC_REGKEY_SYSTEM = "HKEY_LOCAL_MACHINE\Software\Ghisler\Total Commander"
Global Const $TC_REGKEY_USER = "HKEY_CURRENT_USER\Software\Ghisler\Total Commander"
Global Const $TC_SENDMESSAGE = 0x400 + 51 ; $WM_USER + 51

; Environment variables, that can be set at TC start with Autorun
; COMMANDER_LANG - use this language, autodetect intead
; COMMANDER_HWND - main window handle of current instance of TC

; Global constants
Global Const $TC_INI          = EnvGet("COMMANDER_INI")
Global Const $TC_PATH         = __TCHGetPath()
Global Const $TC_HWND         = __TCHGetActiveInstance()
Global Const $TC_LANG         = __TCHGetLangID()
Global Const $TC_SCRIPTS      = $TC_PATH & "\Scripts"
Global Const $TC_CONFIG       = $TC_PATH & "\Profile\Config"
Global Const $TC_CURLANG      = __TCHGetLangID

; Global pseudo-objects, incapsulating language and options handling
Global Const $LANG   = _Lang
Global Const $OPTION = _Option

; #INITIALIZATION# =================================================================================

_Option("ScriptCommonLang",   $TC_SCRIPTS & "\common.lng")
_Option("ScriptCommonIni",    $TC_CONFIG & "\scripts_common.ini")
_Option("DefCmdIconIndex",    0)
_Option("DefScriptIconIndex", 1)
_Option("LangCodes", "ENG|RUS|CZ|DAN|DEU|DUT|ESP|FRA|HUN|ITA|KOR|NOR|POL|ROM|SK|SVN|SWE|CHN")

_IntMsg("Error",          "Error!")
_IntMsg("NoLangId",       "Langauage identifier ""%s"" is not implemented")
_IntMsg("NoInstance",     "Total Commander is not running or error occured.")
_IntMsg("RunError",       "Can't run Total Commander!")
_IntMsg("NotEnoughParam", "Passed %d parameters, should be: %d")
_IntMsg("NoGetPath",      "Can't get the path to Total Commander. Try to run the program from inside of Total Commander.")
_IntMsg("UnnamedScript",  "Unnamed script")
_IntMsg("ErrRemoveTemp",  "Can't remove temporary files/dirs:")
_IntMsg(LangSetFile, _Option("ScriptCommonLang"))

If $TC_LANG <> "ENG" Then _IntMsg(LangInit, $TC_LANG)

Global $g_aGlobalTempList[0]   ; TC temp dirs
OnAutoItExitRegister("__TCHOnExitFunction")

; =============================================================================
; Global pseudo-objects handling
; =============================================================================

; Handle values pairs for options
Func _Option($vID, $vValue = Default)
    Local Static $aList[0][2]
    ; we accept only one function GetList
    If IsFunc($vID) Then Return $vID($aList)
    Local $nIndex = _PairListKeyIndex($aList, $vID)
    If $nIndex = -1 Then
        ; add key/value
        Return $aList[ _PairListAdd($aList, $vID, $vValue) ][1]
    EndIf
    ; get value
    If $vValue = Default Then Return $aList[$nIndex][1]
    ; set value, return old value
    Local $vOld = $aList[$nIndex][1]
    $aList[$nIndex][1] = $vValue
    Return $vOld
EndFunc

; User language pseudo-object
Func _Lang($vID, $vParam1 = Default)
    Local Static $aList[0][2]
    Local Static $sLangFile = Default
    Return _LangHandle($aList, $sLangFile, $vID, $vParam1)
EndFunc

; Internal language pseudo-object
Func _IntMsg($vID, $vParam1 = Default)
    Local Static $aList[0][2]
    Local Static $sLangFile = Default
    Return _LangHandle($aList, $sLangFile, $vID, $vParam1)
EndFunc

; Implementation of handling language pseudo-objects
Func _LangHandle(ByRef $aList, ByRef $sLangFile, $vID, $vParam1 = Default)
    If IsFunc($vID) Then
        Switch FuncName($vID)
            Case "GetList"
                Return $vID($aList)
            Case "LangInit", "LangSetFile"
                Return $vID($aList, $sLangFile, $vParam1)
        EndSwitch
    EndIf
    $nIndex = _PairListKeyIndex($aList, $vID)
    If $nIndex <> -1 Then Return $aList[$nIndex][1]
    If $vParam1 = Default Then Return StringFormat("Not implemented: %s", $vID)
    Return $aList[ _PairListAdd($aList, $vID, $vParam1) ][1]
EndFunc

; Function for setting language file 
Func LangSetFile(ByRef $aList, ByRef $sLangFile, $vParam1 = Default)
    If $vParam1 = Default Then 
        $sLangFile = StringLeft(@ScriptFullPath, StringInStr(@ScriptFullPath, ".", 0, -1)-1) & ".lng"
    Else
        $sLangFile = $vParam1
    EndIf
EndFunc

; Function for initialization 
Func LangInit(ByRef $aList, ByRef $sLangFile, $sLangID = Default)
    If $sLangFile = Default Then LangSetFile($aList, $sLangFile)
    If $sLangID = Default Then $sLangID = $TC_LANG ; if no overriding then use
    If FileExists($sLangFile) Then
        For $i = 0 To UBound($aList)-1
            $aList[$i][1] = IniRead($sLangFile, $sLangID, $aList[$i][0], $aList[$i][1])
        Next
    EndIf
EndFunc

; function for retrieve full list
Func GetList(ByRef $aList)
    Return $aList
EndFunc

; Get key index
Func _PairListKeyIndex(ByRef $aList, $sName)
    For $i = 0 To UBound($aList) - 1
        If $sName = $aList[$i][0] Then Return $i
    Next
    Return -1
EndFunc

; Adds key/value pair
Func _PairListAdd(ByRef $aList, $sName, $vValue = Default)
    Local $nIndex = UBound($aList)
    ReDim $aList[$nIndex+1][2]
    $aList[$nIndex][0] = $sName
    $aList[$nIndex][1] = $vValue
    Return $nIndex
EndFunc

; =============================================================================
; Temporary file/directory handling
; =============================================================================
; #FUNCTION# ==================================================================
; Find free temporary name and return it
; =============================================================================
; Find free temporary name and return it
Func _TCHGetTempFileName($sPrefix = "", $sSuffix = "", $bAutoRemove = True)
    Local $sTemp
    Do
        $sTemp = @TempDir & "\" & $sPrefix & _TCHGetRandomString(8) & $sSuffix
    Until Not FileExists($sTemp)
    
    If $bAutoRemove Then
        ReDim $g_aGlobalTempList[UBound($g_aGlobalTempList)+1]
        $g_aGlobalTempList[UBound($g_aGlobalTempList)-1] = $sTemp
    EndIf

    Return $sTemp
EndFunc

Func _TCHTempDirCreate($sPrefix = "", $sSuffix = "", $bAutoRemove = True)
    Local $sTemp
    Do
        $sTemp = @TempDir & "\" & $sPrefix & _TCHGetRandomString(8) & $sSuffix
    Until Not FileExists($sTemp)
    If DirCreate($sTemp) Then
        If $bAutoRemove Then
            ReDim $g_aGlobalTempList[UBound($g_aGlobalTempList)+1]
            $g_aGlobalTempList[UBound($g_aGlobalTempList)-1] = $sTemp
        EndIf 
        Return $sTemp
    EndIf
    Return SetError(1, 0, "")
EndFunc

; Some functions require script window handle (Sender window for Send functons, context menu etc.)
Func _TCHGetScriptProcHandle()
    Local $sTitle = "ScriptWinRandomTitle" & _TCHGetRandomString(32)
    AutoItWinSetTitle($sTitle)
    Return WinGetHandle($sTitle)
EndFunc

Func _TCHGetRandomString($nLength)
    Local $sString = ""
    For $i = 0 To $nLength - 1
        $sString &= Chr(Random(0x41, 0x5A))
    Next
    Return $sString
EndFunc

Func _TCHGetKeyboardLayout($hWnd = _TCHGetScriptProcHandle())
    Local $ret = DllCall("user32.dll", "long", "GetWindowThreadProcessId", "hwnd", $hWnd, "ptr", 0)
          $ret = DllCall("user32.dll", "long", "GetKeyboardLayout", "long", $ret[0])
    Return "0000" & Hex($ret[0], 4)
EndFunc

Func _TCHSetKeyboardLayout($sLayoutID, $hWnd = _TCHGetScriptProcHandle())
    Local $WM_INPUTLANGCHANGEREQUEST = 0x50
    Local $ret = DllCall("user32.dll", "long", "LoadKeyboardLayout", "str", $sLayoutID, "int", 0)
    DllCall("user32.dll", "int", "SendMessage", "hwnd", $hWnd, _
                                                "int", $WM_INPUTLANGCHANGEREQUEST, _
                                                "int", 1, _
                                                "int", $ret[0])
EndFunc

; copy text with respect of current language layout
Func _TCHCopyClipboard($sText, $sLayoutID = "00000419")
    ; comment copied wrong instead
    _TCHSetKeyboardLayout($sLayoutID)
    ClipPut($sText)
EndFunc

; Due the weird way that TC sends %N param on long files, handle commandline ourself
Func _TCHParseCmdLine($sCmdLine)
    Local $aParam[1] = [0]
    Local $qpos, $spos, $nDouble, $sCurParam
    While StringLen($sCmdLine) > 0
        $sCmdLine = StringStripWS($sCmdLine, 1)
        If StringLeft($sCmdLine, 1) = '"' Then
            $nDouble = 0
            if StringMid($sCmdLine, 2, 1) = '"' Then $nDouble = 1
            $qpos = StringInStr($sCmdLine, '"', 0, 2 + $nDouble)
            $sCurParam = StringMid($sCmdLine, 2 + $nDouble, $qpos - 2 - $nDouble)
            $aParam[0] += 1
            ReDim $aParam[$aParam[0]+1]
            $aParam[$aParam[0]] = $sCurParam
            $sCmdLine = StringTrimLeft($sCmdLine, $qpos + $nDouble)
            If StringLen($sCmdLine) = 1 Then
                ; rare condition that may occur when one redundant trailing qiote exists
                ExitLoop 
            Else
                ContinueLoop
            EndIf
        EndIf
        $spos = StringInStr($sCmdLine, ' ')
        If $spos > 0 Then
            $aParam[0] += 1
            ReDim $aParam[$aParam[0]+1]
            $aParam[$aParam[0]] = StringLeft($sCmdLine, $spos - 1)
            $sCmdLine = StringTrimLeft($sCmdLine, $spos)
        Else
            $aParam[0] += 1
            ReDim $aParam[$aParam[0]+1]
            $aParam[$aParam[0]] = $sCmdLine
            ExitLoop
        EndIf
        
    Wend
    Return $aParam
EndFunc

Func _TCHRestart($hWnd = 0)
    If $hWnd = 0 Then $hWnd = $TC_HWND
    If $hWnd = 0 Then
        MsgBox(48, _IntMsg("Error"), _IntMsg("NoInstance"))
        Exit
    EndIf
    WinClose($hWnd)
    WinWaitClose($hWnd)
    Run($TC_PATH & "\totalcmd.exe") ; 64bit awareness?
    If @error Then MsgBox(48, _IntMsg("Error"), _IntMsg("RunError")))
EndFunc

Func _TCHCheckCmdLine($nReq)
    If $CmdLine[0] < $nReq Then
        MsgBox(48, _IntMsg("Error"), StringFormat(_IntMsg("NotEnoughParam"), $CmdLine[0], $nReq))
        Exit
    EndIf
EndFunc

; ************************************************************************
; Path/filename functions
; Those functions assume that strings that have "\" at the end - are dirs
; ************************************************************************

Func _TCHExtractFileExt($sFile)
    Local $nSPos = StringInStr($sFile, ".", 0, -1)
    If $nSPos > 0 Then
        If Not StringInStr($sFile, "\", 0, 1, $nSPos + 1) Then Return StringMid($sFile, $nSPos + 1)
    EndIf
    Return ""
EndFunc

Func _TCHExtractFileName($sFile, $bNoExtension = False)
    ; check if the file passed is a folder and strip slash
    If StringRight($sFile, 1) = "\" Then $sFile = StringTrimRight($sFile, 1)
    Local $nSPos = StringInStr($sFile, "\", 0, -1)
    If $nSPos > 0 Then $sFile = StringMid($sFile, $nSPos + 1)
    If $bNoExtension And ($sFile <> "") Then
        Local $sExt = _TCHExtractFileExt($sFile)
        If $sExt <> "" Then $sFile = StringTrimRight($sFile, StringLen($sExt) + 1)
    EndIf
    Return $sFile
EndFunc

Func _TCHExtractFilePath($sFile)
    Local $nSPos = StringInStr($sFile, "\", 0, -1)
    If $nSPos = StringLen($sFile) Then
        Return $sFile
    Else
        Return StringLeft($sFile, $nSPos)
    EndIf
EndFunc

Func _TCHChangeExt($sFile, $sNewExt)
    ; beed update to regexp checking
    Local $nSPos = StringInStr($sFile, ".", 0, -1)
    If $nSPos > 0 Then
        Return StringLeft($sFile, $nSPos - 1) & "." & $sNewExt
    EndIf
    Return $sFile
EndFunc

Func _TCHFindTCIconLibrary($bAddEnv = False)
    Local $sIniLib = _TCHIniRead($TC_INI, "Buttonbar", "DefaultLib", "")
    If $sIniLib = "" Then
        $sIconLib = $TC_PATH & "\wciconex."
        If Not (FileExists($sIconLib & "dll") And FileExists($sIconLib & "inc")) Then
            $sIconLib = $TC_PATH & "\wcmicons."
        EndIf
        $sIconLib &= "dll"
    Else
        $sIconLib = $TC_PATH & "\" & $sIniLib & ".dll"
    EndIf
    Return $bAddEnv ? _TCHAddEnv($sIconLib) : $sIconLib
EndFunc

; ************************************************************************
; Window functions
; ************************************************************************

; Standard versions of SendMessage/PostMessage
Func _TCHPostMessage($hWnd, $nMsg, $wParam = 0, $lParam = 0)
	Local $aResult = DllCall("user32.dll", "int", "PostMessage", "hwnd", $hWnd, "int", $nMsg, "wparam", $wParam, "lparam", $lParam)
	If @error Then Return SetError(@error, @extended, "")
	Return $aResult[0]
EndFunc

Func _TCHSendMessage($hWnd, $nMsg, $wParam = 0, $lParam = 0)
	Local $aResult = DllCall("user32.dll", "int", "SendMessage", "hwnd", $hWnd, "int", $nMsg, "wparam", $wParam, "lparam", $lParam)
	If @error Then Return SetError(@error, @extended, "")
	Return $aResult[0]
EndFunc

; Shortcut versions
Func _TCHSendMsg($hWnd, $nMsg)
    Return _TCHSendMessage($hWnd, $TC_SENDMESSAGE, $nMsg, 0)
EndFunc

Func _TCHPostMsg($hWnd, $nMsg)
    Return _TCHPostMessage($hWnd, $TC_SENDMESSAGE, $nMsg, 0)
EndFunc

; Executing user command
Func _TCHExecEMCommand($hTrgWnd, $sCommand, $hSrcWnd = 0)
    Local Const $WM_COPYDATA = 0x004A
    If $hSrcWnd = 0 Then $hSrcWnd = _TCHGetScriptProcHandle()
    Local $pCommand = DllStructCreate("char[" & StringLen($sCommand) & "]")
    DllStructSetData($pCommand, 1, $sCommand)
    Local $pCopyData = DllStructCreate("ulong_ptr;dword;ptr")
    DllStructSetData($pCopyData, 1, Asc('E') + 256 * Asc('M'))
    DllStructSetData($pCopyData, 2, DllStructGetSize($pCommand))
    DllStructSetData($pCopyData, 3, DllStructGetPtr($pCommand))
    DllCall("user32.dll", "int", "SendMessage", "hwnd", $hTrgWnd, "int", $WM_COPYDATA, "wparam", $hSrcWnd, "lparam", DllStructGetPtr($pCopyData))
EndFunc

Func _TCHLockWindowUpdate($hWnd, $bLock)
    _TCHSendMessage($hWnd, 0x000B, $bLock ? 0 : 1) ; WM_SETREDRAW
EndFunc

Func _TCHSetParentModal($hWnd, $bModal = True)
    WinSetState($hWnd, "", $bModal ? @SW_DISABLE : @SW_ENABLE)
EndFunc

Func _TCHTrayMessage($sMessage, $nDelay = 5000)
    TrayTip("Message", $sMessage, 1, 1 + 16)
    Sleep(5000)
EndFunc

; ************************************************************************
; Environment variable functions
; ************************************************************************
; Only first found will be replaced
Func _TCHAddEnv($sText)
    Local $aEnvList[] = ["COMMANDER_PATH","SYSTEMROOT","WINDIR","PROGRAMFILES","COMMONPROGRAMFILES", _
                           "USERPROFILE","ALLUSERSPROFILE","APPDATA"]
    For $i = 0 To UBound($aEnvList) - 1
        $sText = StringReplace($sText, EnvGet($aEnvList[$i]), "%" & $aEnvList[$i] & "%")
        If @extended > 0 Then Return $sText
    Next
    Return $sText
EndFunc

Func _TCHExpandEnv($sText)
    Local $sValue
    $aResult = StringRegExp($sText, "%(\w+)%", 3)
    If IsArray($aResult) Then
        For $i = 0 To UBound($aResult)-1
            $sValue = EnvGet($aResult[$i])
            If $sValue = "" Then ContinueLoop
            $sText = StringReplace($sText, "%" & $aResult[$i] & "%", $sValue)
        Next
    EndIf
    Return $sText
EndFunc

; ************************************************************************
; Misc functions
; ************************************************************************

Func _IsPressed($sHexKey, $vDLL = "user32.dll")
	Local $aReturn = DllCall($vDLL, "short", "GetAsyncKeyState", "int", "0x" & $sHexKey)
	If @error Then Return SetError(@error, @extended, False)
	Return BitAND($aReturn[0], 0x8000) <> 0
EndFunc 

; ************************************************************************
; File functions
; ************************************************************************
Func _FileWriteReplace($sFileName, $sText, $bBackup = False)
    If $bBackup Then FileCopy($sFileName, _TCHExtractFileName($sFileName) & ".bak", 1)
    Local $hFile = FileOpen($sFileName, 2)
    FileWrite($hFile, $sText)
    FileClose($hFile)
EndFunc

; ==================================================================================================
; Continious lists
; ==================================================================================================
; This function read "continious" lists of TC to 2D array
; sKeysFormat - string of key names, separated by "|". As counter substitute using @ char.
; $nStartCount - zero-based index of element from which list is read
; Example for reading WDX plugins list:
; @|@_detect will read keys: 1 and 1_detect, 2 and 2_detect etc.
; ==================================================================================================
Func _TCHIniReadConList($sIniName, $sSection, $sKeysFormat, $nStartCount = 0)
    Local $nNewIdx, $sErrStr = "<<none>>"
    Local $aKeyNames = StringSplit($sKeysFormat, "|")
    Local $nKeys = $aKeyNames[0]
    Local $aResult[1][$nKeys] = [[0]]
    While 1
        $sCurVal = _TCHIniRead($sIniName, $sSection, StringReplace($aKeyNames[1], "@", $aResult[0][0]+$nStartCount), $sErrStr)
        If $sCurVal <> $sErrStr Then
            $aResult[0][0] += 1
            $nNewIdx = $aResult[0][0]
            ReDim $aResult[$nNewIdx+1][$nKeys]
            $aResult[$nNewIdx][0] = $sCurVal
            For $i = 2 To $aKeyNames[0]
                $sCurVal = _TCHIniRead($sIniName, $sSection, StringReplace($aKeyNames[$i], "@", $aResult[0][0]-1+$nStartCount), $sErrStr)
                If $sCurVal <> $sErrStr Then $aResult[$nNewIdx][$i-1] = $sCurVal
            Next
        Else
            Return $aResult
        EndIf
    Wend
EndFunc

; ==================================================================================================
; This function delete item from "continious" ini-lists
; sKeysFormat - string of key names, separated by "|". As counter substitute using @ char.
; $nIndex - zero-based index of element being deleted
; Example for reading WDX plugins list:
; @|@_detect will read keys: 1 and 1_detect, 2 and 2_detect etc.
; ==================================================================================================
Func _TCHIniRemoveConList($sIniName, $sSection, $sKeysFormat, $nIndex)
    Local $aList = _TCHIniReadConList($sIniName, $sSection, $sKeysFormat)
    $nIndex += 1 ; our lists are 1-based, legacy...
    If $nIndex > $aList[0][0] Then Return SetError(1, 0, -1)
    Local $aFields = StringSplit($sKeysFormat, "|", 2)
    Local $nCounter = -1
    For $i = 1 To $aList[0][0]
        If $i = $nIndex Then ContinueLoop ; skip item with index being deleting
        $nCounter += 1
        For $j = 0 To UBound($aFields)-1
            _TCHIniWrite($sIniName, $sSection, StringReplace($aFields[$j], "@", $nCounter), $aList[$i][$j])
        Next
    Next
    ; remove last group
    $nCounter += 1
    For $j = 0 To UBound($aFields)-1
        _TCHIniDelete($sIniName, $sSection, StringReplace($aFields[$j], "@", $nCounter))
    Next
EndFunc

; ==================================================================================================
; Redirection-aware ini functions
; ==================================================================================================

Func _TCHIniGetRedirectName($sIniName, $sSection)
    Local $sErrStr = "<<none>>"
    Local $sBasePath = ""
    Local $sRedirect = IniRead($sIniName, $sSection, "RedirectSection", $sErrStr)
    If $sRedirect <> $sErrStr Then
        If StringInStr($sRedirect, "%") Then $sRedirect = _TCHExpandEnv($sRedirect)
        ; if path is local or UNC then return redirected path itself
        If StringInStr($sRedirect, ":") Or (StringLeft($sRedirect, 2) = "\\")  Then Return $sRedirect
        ; else assume redirect path is relative and extract source ini base path
        $sBasePath = StringRegExpReplace($sRedirect, "(.+\\)([^\\]+)", "\1")
        $sIniName = $sBasePath & $sRedirect
    EndIf
    Return $sIniName
EndFunc

Func _TCHIniDelete($sIniName, $sSection, $sKey = "")
    $sIniName = _TCHIniGetRedirectName($sIniName, $sSection)
    If $sKey = "" Then
        Return IniDelete($sIniName, $sSection)
    Else
        Return IniDelete($sIniName, $sSection, $sKey)
    EndIf
EndFunc

; clear section without deleting it
Func _TCHIniClearSection($sIniName, $sSection)
    $sIniName = _TCHIniGetRedirectName($sIniName, $sSection)
    Local $aKeys = IniReadSection($sIniName, $sSection)
    If UBound($aKeys) = 0 Then Return
    For $i = 1 To $aKeys[0][0]
        IniDelete($sIniName, $sSection, $aKeys[$i][0])
    Next
EndFunc

Func _TCHIniRead($sIniName, $sSection, $sKey, $sDefault = '')
    $sIniName = _TCHIniGetRedirectName($sIniName, $sSection)
    Return IniRead($sIniName, $sSection, $sKey, $sDefault)
EndFunc

Func _TCHIniReadSection($sIniName, $sSection)
    $sIniName = _TCHIniGetRedirectName($sIniName, $sSection)
    Return IniReadSection($sIniName, $sSection)
EndFunc

Func _TCHIniWrite($sIniName, $sSection, $sKey, $sValue)
    $sIniName = _TCHIniGetRedirectName($sIniName, $sSection)
    Return IniWrite($sIniName, $sSection, $sKey, $sValue)
EndFunc

; ==================================================================================================
; Button bar functions
; ==================================================================================================
Func _TCHCreateButton($sCmd, $sParam = "", $sIcon = "", $sTitle = "", $sPath = "")
    Local $sFormat = "TOTALCMD#BAR#DATA\r\n%s\r\n%s\r\n%s\r\n%s\r\n%s\r\n\r\n-1\r\n%s\r\n"
    Return StringFormat($sFormat, $sCmd, $sParam, $sIcon, $sTitle, $sPath)
EndFunc

; save and load backup of "variable" part of buttonbar thar have "command" FIXED_BAR_END or user defined
Func _TCHSaveButtonBarBackup($sBarPath, $sBackupPath, $sFixedBarEndCommand = "FIXED_BAR_END")
    Local $nItemCount = IniRead($sBarPath, "Buttonbar", "Buttoncount", 0)
    Local $bVariable = False
    Local $nCount = 0

    FileDelete($sBackupPath)
    
    For $i = 1 To $nItemCount
        If $bVariable = False Then
            $sCmd = IniRead($sBarPath, "Buttonbar", "cmd" & $i, "")
            If $sCmd = $sFixedBarEndCommand Then
                IniWrite($sBackupPath, "Buttonbar", "Buttoncount", 0) ; stub
                $bVariable = True
            EndIf
            ContinueLoop
        EndIf
        $nCount += 1
        IniWrite($sBackupPath, "Buttonbar", "cmd" & $nCount,    IniRead($sBarPath, "Buttonbar", "cmd" & $i, ""))
        IniWrite($sBackupPath, "Buttonbar", "button" & $nCount, IniRead($sBarPath, "Buttonbar", "button" & $i, ""))
        IniWrite($sBackupPath, "Buttonbar", "iconic" & $nCount, IniRead($sBarPath, "Buttonbar", "iconic" & $i, ""))
        IniWrite($sBackupPath, "Buttonbar", "menu" & $nCount,   IniRead($sBarPath, "Buttonbar", "menu" & $i, ""))
    Next
    If $bVariable Then IniWrite($sBackupPath, "Buttonbar", "Buttoncount", $nCount)
EndFunc

Func _TCHLoadButtonBarBackup($sBarPath, $sBackupPath, $sFixedBarEndCommand = "FIXED_BAR_END")
    Local $nBarItemCount = IniRead($sBarPath, "Buttonbar", "Buttoncount", 0)
    Local $nBackupItemCount = IniRead($sBackupPath, "Buttonbar", "Buttoncount", 0)
    Local $nCount = 0

    For $i = 1 To $nBarItemCount
        $sCmd = IniRead($sBarPath, "Buttonbar", "cmd" & $i, "")
        If $sCmd = $sFixedBarEndCommand Then
            $nCount = $i
            ExitLoop
        EndIf
    Next

    If FileExists($sBackupPath) Then 
        For $i = 1 To $nBackupItemCount
            $nCount += 1
            IniWrite($sBarPath, "Buttonbar", "cmd" & $nCount,    IniRead($sBackupPath, "Buttonbar", "cmd" & $i, ""))
            IniWrite($sBarPath, "Buttonbar", "button" & $nCount, IniRead($sBackupPath, "Buttonbar", "button" & $i, ""))
            IniWrite($sBarPath, "Buttonbar", "iconic" & $nCount, IniRead($sBackupPath, "Buttonbar", "iconic" & $i, ""))
            IniWrite($sBarPath, "Buttonbar", "menu" & $nCount,   IniRead($sBackupPath, "Buttonbar", "menu" & $i, ""))
        Next
        
        IniWrite($sBarPath, "Buttonbar", "Buttoncount", $nCount)
    EndIf
EndFunc

; ==================================================================================================
; Internal helper functions
; ==================================================================================================

; Get active instance of Total Commander (in order multiple copies runned)
Func __TCHGetActiveInstance()
    Local $sInst = EnvGet("COMMANDER_HWND") ; if set by Autorun
    If ($sInst <> "") And StringIsDigit($sInst) Then Return Hwnd($sInst)
    Local $aList = WinList("[CLASS:TTOTAL_CMD]")
    For $i = 1 To $aList[0][0]
        If WinActive($aList[$i][1]) Then Return $aList[$i][1]
    Next
    Return 0
EndFunc

; Try to get install path of TC (in priority: EnvVar, HKLM key, HKCU key)
Func __TCHGetPath()
    Local $sTCPath = EnvGet("COMMANDER_PATH")
    If $sTCPath = "" Then ; unlikely, legacy code below
        $sTCPath = RegRead($TC_REGKEY_SYSTEM, "InstallDir")
        If $sTCPath = "" Then
            $sTCPath = RegRead($TC_REGKEY_USER, "InstallDir")
            If $sTCPath = "" Then
                MsgBox(48, _IntMsg("Error"), _IntMsg("NoGetPath"))
                Exit
            EndIf
        EndIf
    EndIf
    Return $sTCPath
EndFunc

Func __TCHGetLangID()
    Local $sLng = EnvGet("COMMANDER_LANG") ; if set by Autorun
    If $sLng = "" Then
        $sLng = _TCHIniRead($TC_INI, "Configuration", "LanguageIni", "WCMD_ENG.LNG")
        $sLng = StringRegExpReplace($sLng, "(?i).+[_-]([a-zA-Z]+)\.lng", "\1")
    EndIf
    Return $sLng
EndFunc

Func __TCHOnExitFunction()
    Local $sList = ""
    For $vItem In $g_aGlobalTempList
        If StringInStr(FileGetAttrib($vItem), "D") Then
            FileChangeDir(@ScriptDir) ; get out of temp dir
            If Not DirRemove($vItem, 1) Then $sList &= $vItem & @CRLF
        Else
            If Not FileDelete($vItem) Then $sList &= $vItem & @CRLF
        EndIf
    Next
    If $sList <> "" Then MsgBox(48, _IntMsg("Error"), _IntMsg("ErrRemoveTemp") & @CRLF & $sList)
EndFunc