#cs =============================================================================================================================
[info]
Param=%WF "%P"
Icon=%COMMANDER_PATH%\Icons\Scripts.icl,0
Info.ENG.Title=Comment editing
Info.ENG.Description=Editing comments for multiple files at once
Info.RUS.Title= 
Info.RUS.Description=     
Info.META.Version=2.4b
Info.META.Date=2020-06-15
Info.META.Author=Dmitry Yudin
[/info]
#ce =============================================================================================================================

#include <TC\TCHelpers.au3>
#include <TC\DescrIon.au3>
#include <TC\TCGUIHelpers.au3>

#include <WinAPIvkeysConstants.au3>
#include <Common\HotKey.au3>

#include <ButtonConstants.au3>
#include <EditConstants.au3>
#include <ComboConstants.au3>
#include <StaticConstants.au3>
#include <GUIConstantsEx.au3>
#include <WindowsConstants.au3>
#include <ListViewConstants.au3>

#include <Array.au3>

#NoTrayIcon

_TCHCheckCmdLine(2)

;$t = TimerInit()


Global $hParent = $TC_HWND
Global $sFlags = $CmdLine[0] >= 3 ? $CmdLine[3] : ""

Global $sFile = FileRead($CmdLine[1])
Global $aList = StringSplit($sFile, @CRLF, 3)

; TC save list with empty line at the end, need to remove it
If UBound($aList) > 0 And $aList[UBound($aList)-1] = "" Then
    ReDim $aList[UBound($aList)-1]
EndIf

; Only single file is selected and MCE_MULTI_ONLY set, invoke standard function
If (UBound($aList) < 2) And StringInStr($sFlags, "MCE_MULTI_ONLY") Then
     _TCHPostMsg($hParent, 2700) ; Post "cm_EditComment" message to active TC instance
    Exit
EndIf

OnAutoItExitRegister ("_OnExit")

; folders should not have "\" at the end
For $i = 0 To UBound($aList)-1
    If StringRight($aList[$i], 1) = "\" Then $aList[$i] = StringTrimRight($aList[$i], 1)
Next

Global $sDescrIon = $CmdLine[2] & "descript.ion"

Global $nDefEnc = $ENC_UTF8
Global $sCurrFile = ""

; global flag, true if names OR existing descript.ion has unicode chars
Global $bNeedUnicode = _StringHasUniChars($sFile)

Global $bDisableSingleEdit = False

If $sFlags <> "" Then
    Global $aFlags = StringSplit($CmdLine[3], "|")
    For $i = 1 To $aFlags[0]
        Switch $aFlags[$i]
            Case "MCE_DEFENC_UTF8"
                $nDefEnc = $ENC_UTF8
            Case "MCE_DEFENC_UTF16"
                $nDefEnc = $ENC_UTF16
            Case "MCE_DEFENC_UTF16BE"
                $nDefEnc = $ENC_UTF16BE
            Case "MCE_DISABLE_SINGLEEDIT"
                $bDisableSingleEdit = True
        EndSwitch
    Next
EndIf

Global $nCurEnc, $nNewEnc

$LANG("1",  "File Comment")
$LANG("2",  "MULTIEDIT: comments for %d files")
$LANG("3",  "Options")
$LANG("4",  "Replace")
$LANG("5",  "Add")
$LANG("6",  "Beginning")
$LANG("7",  "End")
$LANG("8",  "Cancel")
$LANG("9",  "Neighbour")
$LANG("10", "Neighbour descriptions")
$LANG("11", "Edit comment for:")
$LANG("12", "Can't create comment on FTP.")
$LANG("13", "Can't create comment in plugin.")
$LANG("14", "Base path is empty (plugin?).")
$LANG("15", "Encoding:")
$LANG("16", "Warning")
$LANG("17", "Changing encoding from unicode to ANSI may cause data loss!")
$LANG("18", "When changing from ANSI to unicode, file names may need to be converted. Do you want the program try to do this?")
$LANG("19", "None")
$LANG(LangInit)

; Check correctness of base path (not all variants can be checked...)
Select
    Case StringRegExp($CmdLine[2], "^/.+") ; FTP
        MsgBox(0, "Error", $LANG("12"))
        Exit
    Case StringRegExp($CmdLine[2], "^\\{3}[^\\]+") ; Plugin
        MsgBox(0, "Error", $LANG("13"))
        Exit
    Case $cmdline[2] = ""
        MsgBox(0, "Error", $LANG("14"))
        Exit
EndSelect


Global $aDescr = _DescrIon_Load($sDescrIon)
If @extended = 1 Then ; created new
    $nCurEnc = 0
    $nNewEnc = $nDefEnc
Else
    $nCurEnc = _DescrIon_GetMeta($aDescr, $DI_META_ENCODING)
    $nNewEnc = $nCurEnc
EndIf

Global $sDefDescr = _GetDefDescr($aDescr, $aList)

If Not $bNeedUnicode Then $bNeedUnicode = _DescrIon_GetMeta($aDescr, $DI_META_HASUNICHARS)

_UpdateEncState($bNeedUnicode)

Opt("GUIResizeMode", $GUI_DOCKAUTO)

$hGUI = GUICreate($LANG("1"), 625, 313, -1, -1, BitOR($WS_POPUP, $WS_CAPTION, $WS_SYSMENU), $WS_EX_COMPOSITED, $hParent)
$hDummy = GUICtrlCreateDummy()
$hDummyContext = GUICtrlCreateContextMenu($hDummy)
$hReplace = GUICtrlCreateMenuItem($LANG("4"), $hDummyContext, -1 , 1)
GUICtrlSetState(-1, $GUI_CHECKED)
$hAdd = GUICtrlCreateMenuItem($LANG("5"), $hDummyContext, -1 , 1)
$hSep = GUICtrlCreateMenuItem("", $hDummyContext)
$hBeginning = GUICtrlCreateMenuItem($LANG("6"), $hDummyContext, -1 , 1)
$hEnd = GUICtrlCreateMenuItem($LANG("7"), $hDummyContext, -1 , 1)
GUICtrlSetState(-1, $GUI_CHECKED)
$hCancel = GUICtrlCreateButton($LANG("8"), 536, 283, 80, 24, 0)
$hOK = GUICtrlCreateButton("F2 OK", 448, 283, 80, 24, $BS_DEFPUSHBUTTON)
$hOptions = GUICtrlCreateButton($LANG("3"), 8, 283, 75, 24, $BS_NOTIFY)
GUICtrlCreateLabel($LANG("15"), 400, 8, 70, 17, $SS_CENTER)
GUICtrlCreateLabel(_GetEncString($nCurEnc), 470, 8, 50, 17, $SS_CENTER)
GUICtrlSetFont(-1, -1, 800)
GUICtrlCreateLabel(">>", 520, 8, 20, 17)
$hEncSwitch = GUICtrlCreateCombo("", 540, 4, 80, 24, BitOR($CBS_DROPDOWNLIST, $CBS_AUTOHSCROLL))
GUICtrlSetData($hEncSwitch, _GetEncComboString($bNeedUnicode))  ;"ANSI|UTF-16|UTF-16BE|UTF-8")

If UBound($aList) > 1 Then
    $hEdit = GUICtrlCreateEdit("", 4, 32, 616, 244, BitOR($ES_WANTRETURN,$WS_VSCROLL))
    If $bDisableSingleEdit Then
        $hEditComm = GUICtrlCreateLabel(StringFormat($LANG("2"), UBound($aList)), 8, 8, 400, 17, $DT_PATH_ELLIPSIS)
    Else
        $hEditComm = GUICtrlCreateCombo("", 4, 6, 380, 17, BitOR($CBS_DROPDOWNLIST, $WS_VSCROLL))
        GUICtrlSetData($hEditComm, _FileListToString($aList))
        GUICtrlSendMsg($hEditComm, $CB_SETCURSEL , 0, 0)
    EndIf
Else
    $sCurrFile = $aList[0]
    $hEdit = GUICtrlCreateEdit("", 4, 40, 616, 236, BitOR($ES_WANTRETURN,$WS_VSCROLL))
    GUICtrlCreateLabel($LANG("11"), 8, 4, 296, 17, 0)
    $hEditComm = GUICtrlCreateLabel($sCurrFile, 8, 22, 400, 17, $DT_PATH_ELLIPSIS)
    GUICtrlSetTip($hEditComm, $sCurrFile)
    GUICtrlSetState($hOptions, $GUI_DISABLE)
EndIf

$hSize = GUICtrlCreateLabel("", 269, 289, 84, 17, $SS_CENTER)
$hOther = GUICtrlCreateButton($LANG("9"), 90, 283, 75, 24)
Dim $hGUI_AccelTable[1][2] = [["{F2}", $hOK]]
GUISetAccelerators($hGUI_AccelTable)

If $aDescr[0][0] < 1 Then GUICtrlSetState($hOther, $GUI_DISABLE)

_SetEncCombo($nNewEnc, $bNeedUnicode)

GUIRegisterMsg($WM_COMMAND, "WM_COMMAND_FUNC")
GUICtrlSetData($hEdit, $sDefDescr)
GUICtrlSetData($hSize, StringLen($sDefDescr))

; apply default description, or it will be lost when files switching
_DescrIon_SetDescr($aDescr, $sCurrFile, $sDefDescr)

SetFocus(GUICtrlGetHandle($hEdit))

_TCH_GUI_SetDefaultIcon()

;_ArrayDisplay($aDescr)

GUISetState(@SW_SHOW)

_TCH_GUI_AdjustSize($hGUI)

; disable total commander window
_TCHSetParentModal($hParent)

; assign required hotkeys
_HotKeyAssign(BitOR($CK_CONTROL, $VK_A), '_SelectAll')
_HotKeyAssign(BitOR($CK_CONTROL, $VK_W), '_OpenOtherWin')
_HotKeyAssign(BitOR($CK_CONTROL, $VK_S), '_DoneEdit')
_HotKeyAssign(BitOR($CK_ALT, $VK_UP),    '_PrevFile')
_HotKeyAssign(BitOR($CK_ALT, $VK_DOWN),  '_NextFile')
_HotKeyAssign(BitOR($CK_ALT, $VK_HOME),  '_MultiFile')

;MsgBox(0, "title", TimerDiff($t))


While 1
    $nMsg = GUIGetMsg()
    Switch $nMsg
        Case $hDummy
            $nMsg = GUICtrlRead($hDummy)
            ContinueCase
        Case $GUI_EVENT_CLOSE, $hCancel
            Exit
        Case $hEditComm
            $nIndex = GUICtrlSendMsg($hEditComm, $CB_GETCURSEL, 0, 0)
            _SwitchFile($nIndex)
        Case $hEncSwitch
            $nNewEnc = _GetEncCombo(_UnicodeState())
#cs
            $nNewEncoding = GUICtrlSendMsg($hEncSwitch, $CB_GETCURSEL, 0, 0) + 1
            If ($nNewEncoding = 1) And ($g_nOldEncoding > $nNewEncoding) Then
                MsgBox(48, $LANG("16"), $LANG("17"), 0, $hGUI)
            ElseIf ($nNewEncoding > 1) And ($g_nOldEncoding = 1) And ($aDescr[0][0] > 0) Then
                If MsgBox(64 + 4, "", $LANG("18"), 0, $hGUI) = 6 Then 
                    For $i = 0 To UBound($aList)-1
                        _DescrIon_ReplaceAnsiName($aDescr, $aList[$i])
                    Next
                EndIf
            EndIf
            _DescrIon_SetMeta($aDescr, $DI_META_ENCODING, $nNewEncoding)
#ce
        Case $hOK, $hDummy
            $nNewEnc = _GetEncCombo(_UnicodeState())
            _DescrIon_SetMeta($aDescr, $DI_META_ENCODING, $nNewEnc)
            _DoSave()
            ;GUISetState(@SW_HIDE, $hGUI)
            ExitLoop
        Case $hOptions
            If BitAND(GUICtrlRead($hReplace), $GUI_CHECKED) Then
                GUICtrlSetState($hBeginning, $GUI_DISABLE)
                GUICtrlSetState($hEnd, $GUI_DISABLE)
            Else
                GUICtrlSetState($hBeginning, $GUI_ENABLE)
                GUICtrlSetState($hEnd, $GUI_ENABLE)
            EndIf
            $pos = ControlGetPos($hGUI, "", $hOptions)
            _ClientToScreen($hGUI, $pos[0], $pos[1])
            _TrackPopupMenu($hGUI, GUICtrlGetHandle($hDummyContext), $pos[0], $pos[1]+$pos[3])
        Case $hOther
            _OpenOtherWin()
    EndSwitch
WEnd


Func _DoSave()
    Local $sNewDescr = GUICtrlRead($hEdit)

    ; if we upgrade from the ANSI to Unicode, and there was broken names, try to find and replace them
    If _DescrIon_NeedNameCheck($aDescr) And ($nNewEnc <> $ENC_ANSI) Then
        For $i = 0 To UBound($aList)-1
            _DescrIon_ReplaceAnsiName($aDescr, $aList[$i])
        Next
    EndIf

    If (UBound($aList) > 1) Then
        $nIndex = GUICtrlSendMsg($hEditComm, $CB_GETCURSEL, 0, 0)
        If ($nIndex = 0) Then ; chosen multicomment
            For $i = 0 To UBound($aList)-1
                ; if we by mistake selected descript.ion itself
                If $aList[$i] = $sDescrIon Then ContinueLoop
                If BitAND(GUICtrlRead($hReplace), $GUI_CHECKED) Then
                    _DescrIon_SetDescr($aDescr, $aList[$i], $sNewDescr)
                Else
                    $sCurDescr = _DescrIon_GetDescr($aDescr, $aList[$i])
                    If BitAND(GUICtrlRead($hBeginning), $GUI_CHECKED) Then
                        _DescrIon_SetDescr($aDescr, $aList[$i], $sNewDescr & $sCurDescr)
                    Else
                        _DescrIon_SetDescr($aDescr, $aList[$i], $sCurDescr & $sNewDescr)
                    EndIf
                EndIf
            Next
        EndIf
    EndIf

    _DescrIon_Save($sDescrIon, $aDescr)
EndFunc 

;=================================================================================================================================
;=================== Functions ===================================================================================================
;=================================================================================================================================

Func _UpdateEncState($bState)
    If $bState And $nNewEnc = $ENC_ANSI Then $nNewEnc = $nDefEnc
EndFunc

Func _GetEncCombo($bUnicode)
    Return GUICtrlSendMsg($hEncSwitch, $CB_GETCURSEL, 0, 0) + ($bUnicode ? 2 : 1)
EndFunc

Func _SetEncCombo($nCurEnc, $bUnicode)
    GUICtrlSendMsg($hEncSwitch, $CB_SETCURSEL , $nNewEnc - ($bUnicode ? 2 : 1), 0)
EndFunc

Func _SetEncComboState($bUnicode)
    Local $nCount = GUICtrlSendMsg($hEncSwitch, $CB_GETCOUNT, 0, 0)    
    If $bUnicode Then
        If $nCount = 3 Then Return
    Else
        If $nCount = 4 Then Return
    EndIf
    GUICtrlSetData($hEncSwitch, "|" & _GetEncComboString($bUnicode))
    _SetEncCombo($nNewEnc, $bUnicode)
EndFunc

Func _GetEncString($nEnc)
    Local Static $aEnc[] = [$LANG("19"),"ANSI","UTF-16","UTF-16BE","UTF-8"]
    Return $aEnc[$nEnc]
EndFunc

Func _GetEncComboString($bUnicode)
    Local $sUnicode = "UTF-16|UTF-16BE|UTF-8"
    Return $bUnicode ? $sUnicode : "ANSI|" & $sUnicode
EndFunc

Func _SwitchFile($nIndex)
    If $nIndex = 0 Then 
        $sCurrFile = ""
        GUICtrlSetTip($hEditComm, "")
        GUICtrlSetState($hOptions, $GUI_ENABLE)
;        GUICtrlSetColor($hEditComm, 0xDD0000)
    Else
        $sCurrFile = GUICtrlRead($hEditComm)
        GUICtrlSetTip($hEditComm, $sCurrFile)
        GUICtrlSetState($hOptions, $GUI_DISABLE)
;        GUICtrlSetColor($hEditComm, 0x0)
    EndIf

    $sCurDescr = _DescrIon_GetDescr($aDescr, $sCurrFile)

    GUICtrlSetData($hEdit, $sCurDescr)
    GUICtrlSetData($hSize, StringLen($sCurDescr))
    SetFocus(GUICtrlGetHandle($hEdit))

    ; when mode changed, check if it's need to be unicode and switch
    Local $bState = _UnicodeState($sCurrFile, $sCurDescr)
    _UpdateEncState($bState)
    _SetEncComboState($bState)
EndFunc

Func _FileListToString(ByRef $aList)
    Local $sText = StringFormat($LANG("2"), UBound($aList)) & "|"
    For $i = 0 To UBound($aList)-1
        $sText &= $aList[$i] & "|"
    Next
    Return StringTrimRight($sText, 1)
EndFunc


Func _GetOtherDescr($hWnd)
    Local $aUniqueDescr[1] = [0], $selected, $data = "", $nMsg

    Local $hOtherGUI = GUICreate($LANG("10"), 564, 265, -1, -1, BitOR($WS_POPUP, $WS_CAPTION, $WS_SYSMENU), -1, $hWnd)
    Global $hList = GUICtrlCreateListView("Description", 8, 8, 548, 214, BitOR($GUI_SS_DEFAULT_LISTVIEW, $LVS_NOCOLUMNHEADER), BitOR($WS_EX_CLIENTEDGE,$LVS_EX_FULLROWSELECT))
    GUICtrlSendMsg(-1, $LVM_SETCOLUMNWIDTH, 0, -2)
    Local $hOK = GUICtrlCreateButton("OK", 212, 232, 75, 25, $BS_DEFPUSHBUTTON)
    Local $hOtherCancel = GUICtrlCreateButton($LANG("8"), 292, 232, 75, 25, 0)

    _TCH_GUI_SetDefaultIcon()

    For $i=1 To $aDescr[0][0]
        For $j = 1 To $aUniqueDescr[0]
            If $aUniqueDescr[$j] = $aDescr[$i][1] Then ContinueLoop(2)
        Next
        $aUniqueDescr[0] += 1
        ReDim $aUniqueDescr[$aUniqueDescr[0]+1]
        $aUniqueDescr[$aUniqueDescr[0]] = $aDescr[$i][1]
    Next

    _ArraySort($aUniqueDescr, 0, 1)

    Local $ahListDescr[$aUniqueDescr[0]+1][2] = [[$aUniqueDescr[0], 0]]

    For $i = 1 To $aUniqueDescr[0]
        $ahListDescr[$i][0] = $aUniqueDescr[$i]
        $ahListDescr[$i][1] = GUICtrlCreateListViewItem(StringReplace($aUniqueDescr[$i], @LF, "  "), $hList)
    Next

    GUISetState(@SW_SHOW)
    GUISwitch($hOtherGUI)
    GUISetState(@SW_DISABLE, $hWnd)

    _TCH_GUI_AdjustSize($hOtherGUI)

    _TCHG_DoubleClickHandlerAdd($hList, $hOK)
    _TCHG_MessageHandlerRegister($hOtherGUI)
    
    DllCall("user32.dll", "int", "ShowScrollBar", "hwnd", GUICtrlGetHandle($hList), "int", 0, "int", False)

    While 1
	    $nMsg = GUIGetMsg()
        Switch $nMsg
            Case $GUI_EVENT_CLOSE, $hOtherCancel
                ExitLoop
            Case $hOK
                $selected = GUICtrlRead($hList)
                $data = ""
                If $selected > 0 Then
                    For $i = 1 To $ahListDescr[0][0]
                        If $ahListDescr[$i][1] = $selected Then
                            $data = $ahListDescr[$i][0]
                            ExitLoop
                        EndIf
                    Next
                    ExitLoop
                EndIf
        EndSwitch
    Wend

     _TCHG_MessageHandlerUnregister($hOtherGUI)

    GUISetState(@SW_ENABLE, $hWnd)
    GUIDelete($hOtherGUI)
    GUISwitch($hWnd)
    Return StringReplace($data, @LF, @CRLF)
EndFunc

; handler for counting edit length
Func WM_COMMAND_FUNC($hWnd, $msg, $wParam, $lParam)
    Local $nNotifyCode = BitShift($wParam, 16)
    Local $nID = BitAND($wParam, 0xFFFF)
    Local $nMsg = BitShift($wParam, 16)
    If ($nID = $hEdit) And ($nMsg = $EN_CHANGE) Then
        Local $sText = GUICtrlRead($hEdit)
        Local $bState = _UnicodeState($sCurrFile, $sText)
        GUICtrlSetData($hSize, StringLen($sText))
        _UpdateEncState($bState)
        _SetEncComboState($bState)
        _DescrIon_SetDescr($aDescr, $sCurrFile, $sText)
    EndIf 
    Return $GUI_RUNDEFMSG
EndFunc

; founds if description for all selected files are the same and return it
Func _GetDefDescr1(ByRef $aDescr, ByRef $aList)
    Local $sComm = $mTest[$aList[0]]
    For $i = 1 To UBound($aList) - 1
        If Not ($sComm = $mTest[$aList[$i]]) Then Return ""
    Next
    Return $sComm
EndFunc

Func _GetDefDescr(ByRef $aDescr, ByRef $aList)
    Local $sComm = _DescrIon_GetDescr($aDescr, $aList[0])
    ; if first comment is empty, it's not matter if any other will have text or not 
    If $sComm = "" Then Return ""
    For $i = 1 To UBound($aList) - 1
        If Not ($sComm = _DescrIon_GetDescr($aDescr, $aList[$i])) Then Return ""
    Next
    Return $sComm
EndFunc

Func _UnicodeState($sFile = Default, $sText = Default)
    Local Static $aList[1] = [0]
    ; global state is "Need" - always True
    If $bNeedUnicode Then Return True
    ; if multifile mode - check only it's text
    If $sFile = "" Then Return _StringHasUniChars($sText)
    ; next handle the single file mode
    If $sFile = Default And $aList[0] = 0 Then Return False
    For $i = 1 To $aList[0]
        If $aList[$i] = $sFile Then ExitLoop
    Next
    Local $bHasChars = _StringHasUniChars($sText)
    If $i > $aList[0] Then
        If Not $bHasChars Then Return $aList[0] <> 0
        $aList[0] += 1
        ReDim $aList[$aList[0]+1]
        $aList[$aList[0]] = $sFile
        Return True
    Else
       If $bHasChars Then Return True
       If $i < $aList[0] Then $aList[$i] = $aList[$aList[0]]
       ReDim $aList[$aList[0]]
       $aList[0] -= 1
       Return $aList[0] <> 0
    EndIf
EndFunc

Func _OnExit()
    _TCHSetParentModal($hParent, False)
EndFunc

;=================================================================================================================================
; Hotkey handlers
;=================================================================================================================================

Func _OpenOtherWin()
    ; don't open dialog if no neighbours
    If BitAND(GUICtrlGetState($hOther), $GUI_DISABLE) Then Return
    Local $sNewDescr = _GetOtherDescr($hGUI)
    If $sNewDescr <> "" Then
        _DescrIon_SetDescr($aDescr, $sCurrFile, $sNewDescr)
        GUICtrlSetData($hEdit, $sNewDescr)
        Local $nLen = StringLen($sNewDescr)
        GUICtrlSetData($hSize, $nLen)
        GUICtrlSendMsg($hEdit, $EM_SETSEL, $nLen, $nLen)
        SetFocus(GUICtrlGetHandle($hEdit))
    Else
        GUICtrlSetState($hEdit, $GUI_FOCUS)
    EndIf
EndFunc

Func _DoneEdit()
    GUICtrlSendMsg($hOK, $BM_CLICK, 0, 0)
EndFunc

Func _SelectAll()
    GUICtrlSendMsg($hEdit, $EM_SETSEL, 0, -1)
EndFunc

Func _PrevFile()
    Local $nIndex = GUICtrlSendMsg($hEditComm, $CB_GETCURSEL, 0, 0)
    If $nIndex = 0 Then Return
    $nIndex -= 1
    GUICtrlSendMsg($hEditComm, $CB_SETCURSEL, $nIndex, 0)
    _SwitchFile($nIndex)
EndFunc

Func _NextFile()
    Local $nIndex = GUICtrlSendMsg($hEditComm, $CB_GETCURSEL, 0, 0)
    Local $nCount = GUICtrlSendMsg($hEditComm, $CB_GETCOUNT, 0, 0)    
    If $nIndex = $nCount Then Return
    $nIndex += 1
    GUICtrlSendMsg($hEditComm, $CB_SETCURSEL, $nIndex, 0)
    _SwitchFile($nIndex)
EndFunc

Func _MultiFile()
    GUICtrlSendMsg($hEditComm, $CB_SETCURSEL, 0, 0)
    _SwitchFile(0)
EndFunc

;=================================================================================================================================
; Helpers
;=================================================================================================================================

; when using standard GUICtrlSetState($hEdit, $GUI_FOCUS), cursor position is
; always resets to line start (bug?), so use API function
Func SetFocus($hWnd)
    DllCall("user32.dll", "hwnd", "SetFocus", "hwnd", $hWnd)
EndFunc